class SerializableDataclassRegistry(object):
    __instance = None
    __registered_classes = {}

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(SerializableDataclassRegistry, cls).__new__(cls)
        return cls.__instance

    def register(self, cls: type):
        self.__registered_classes[cls.__qualname__] = cls

    def __len__(self):
        return len(self.__registered_classes)



from .consts import Lang
from .errors import DeserializeFailedError, DeserializeConfusedError
from .entry import serializable_dataclass

__all__ = [
    serializable_dataclass,
    Lang,
    DeserializeFailedError,
    DeserializeConfusedError,
]



BASIC_TYPES = [int, float, complex, bool, str, type(None)]


def _is_union_type(typing):
    # Python3.8 does not support GenericAlias yet
    if getattr(typing, "__origin__", False):
        type_name = getattr(typing.__origin__, "_name", None)

        return type_name == "Union"

    return False


def _is_sequence_type(typing):
    return getattr(typing, "_name") == "Sequence"


def _is_list_type(typing):
    return getattr(typing, "_name") == "List"


def _is_tuple_type(typing):
    return getattr(typing, "_name") == "Tuple"


def _is_set_type(typing):
    return getattr(typing, "_name") == "Set"



from decimal import Decimal
from enum import Enum
from types import GeneratorType
from typing import Text, Dict, Union, Sequence, List, Any

from .consts import _SERIALIZABLE_SIGN
from .errors import DeserializeConfusedError, DeserializeFailedError
from .type_system import (
    BASIC_TYPES,
    _is_union_type,
    _is_sequence_type,
    _is_list_type,
    _is_tuple_type,
    _is_set_type,
)


def __serialize_value(value: Any):
    if getattr(value, _SERIALIZABLE_SIGN, False):
        return value.serialize()
    elif isinstance(value, Decimal):
        return str(value)
    elif isinstance(value, Enum):
        return value.value
    elif (
        isinstance(value, list)
        or isinstance(value, set)
        or isinstance(value, tuple)
        or isinstance(value, GeneratorType)
    ):
        return [__serialize_value(v) for v in value]
    elif isinstance(value, dict):
        return {str(k): __serialize_value(value[k]) for k in value}
    else:
        return value


def json_serializer(self):
    cls_annotations = self.__annotations__
    result = {}

    for field_name in cls_annotations:
        value = getattr(self, field_name)
        result[field_name] = __serialize_value(value)

    return result


def __deserialize_value_for_all_assumed_types(value: Any, desired_types: Sequence[type]):
    hashable_possible_values = set()
    unhashable_possible_values = []
    possible_values_len = lambda: len(hashable_possible_values) + len(unhashable_possible_values)

    for desired_type in desired_types:
        possible_value = __deserialize_value(value, desired_type)
        try:
            hashable_possible_values.add(possible_value)
        except TypeError:
            unhashable_possible_values.append(possible_value)

    if possible_values_len() > 1:
        raise DeserializeConfusedError(
            f"more than 1 possible deserialized value for {value}: {hashable_possible_values}"
        )
    if possible_values_len() == 0:
        raise DeserializeFailedError(
            f"could not deserialize value {value} under types {desired_types}"
        )

    if len(hashable_possible_values) > 0:
        return hashable_possible_values.pop()
    else:
        return unhashable_possible_values[0]


def __deserialize_value(value: Any, desired_type: type):
    if desired_type in BASIC_TYPES:
        return value

    # for deserializable_dataclass type
    if getattr(desired_type, _SERIALIZABLE_SIGN, False):
        return desired_type.deserialize(value)

    if desired_type is Decimal:
        return Decimal(value)

    if hasattr(desired_type, "mro") and Enum in desired_type.mro():
        return desired_type(value)

    # for sequence type
    if _is_sequence_type(desired_type) or _is_list_type(desired_type):
        return [__deserialize_value_for_all_assumed_types(v, desired_type.__args__) for v in value]

    if _is_tuple_type(desired_type):
        return tuple(
            __deserialize_value_for_all_assumed_types(v, desired_type.__args__) for v in value
        )

    if _is_set_type(desired_type):
        return {__deserialize_value_for_all_assumed_types(v, desired_type.__args__) for v in value}

    # for Generic types (Union, Option)
    if _is_union_type(desired_type):
        type_args = desired_type.__args__
        return __deserialize_value_for_all_assumed_types(value, type_args)

    raise NotImplementedError(f"deserialize {value} for type {desired_type} is not supported yet")


@classmethod
def json_deserializer(cls, json: Dict[Text, Union[int, complex, bool, List, Sequence, Dict]]):
    cls_annotations = cls.__annotations__
    obj_values = {}

    for field_name in cls_annotations:
        desired_type = cls_annotations[field_name]
        field_value = json.get(field_name)
        obj_values[field_name] = __deserialize_value(field_value, desired_type)

    return cls(**obj_values)



from dataclasses import _process_class

from .consts import Lang, _SERIALIZABLE_SIGN
from .json import json_serializer, json_deserializer
from .registry import SerializableDataclassRegistry


def serializable_dataclass(
    cls=None,
    *,
    language=Lang.JSON,
    init=True,
    repr=True,
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,
):
    def wrapper(cls):
        processed_class = _process_class(
            cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash, frozen=frozen
        )
        if language == Lang.JSON:
            serialize_fn = json_serializer
            deserialize_fn = json_deserializer
        else:
            raise NotImplementedError(f"language {language} does not supported")
        setattr(processed_class, _SERIALIZABLE_SIGN, True)
        setattr(processed_class, "serialize", serialize_fn)
        setattr(processed_class, "deserialize", deserialize_fn)
        return processed_class

    # call with no parameters
    if cls is None:
        return wrapper

    SerializableDataclassRegistry().register(cls)

    # class with parameters
    return wrapper(cls)



import argparse
import re
import subprocess

version_pattern = r"\d\.\d\.\d"
parser = argparse.ArgumentParser()
parser.add_argument("version", help="a SEMVER string X.Y.Z")
args = parser.parse_args()
if not re.match(version_pattern, args.version):
    print("argument must be SEMVER string in format X.Y.Z")
else:
    with open("setup.py") as fp:
        old_setupfile = fp.read()
    new_setupfile = re.sub(
        f"version='{version_pattern}'", f"version='{args.version}'", old_setupfile
    )
    with open("setup.py", "w") as fp:
        print(new_setupfile, file=fp)

    subprocess.run(["./publish.sh", "v" + args.version])
